/*
 * generatetrees.c
 *
 * ($ID$)
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <ctype.h>

int Maxspc;
int treestrlen;
char *treestr;
char **spcnames;
FILE *outfile;
int  dotnum = 0;


/***************************************************************/

/*
 *   prints characters according to counterlevels:
 *
 *   e.g. printdots(stdout, winnum, 10, '.', 100, ' ', 500, '\n');
 *
 *   increments winnum, and prints '.' if winnum is multiple of 10,
 *   and space if winnum is multiple of 100, and LF if winnum is 
 *   multiple of 500 to stdout.
 */
void printdots (FILE *output, 
                int  *counter, 
                int   th1,
                char  ch1,
                int   th2,
                char  ch2,
                int   th3,
                char  ch3)
{ /* printdots */
    (*counter)++; 
    if ((*counter % th1) == 0) {
       fprintf(output,"%c", ch1); 
       fflush(output);
       if ((*counter % th2) == 0) fprintf(output,"%c",ch2);
       if ((*counter % th3) == 0) fprintf(output,"%c",ch3);
    }
} /* printdots */


/***************************************************************/


/******************************************************************************/
/* random numbers generator  (Numerical recipes)                              */
/******************************************************************************/

/* definitions */
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

/* variable */
long idum;

double randomunitintervall()
/* Long period (> 2e18) random number generator. Returns a uniform random
   deviate between 0.0 and 1.0 (exclusive of endpoint values).

   Source:
   Press et al., "Numerical recipes in C", Cambridge University Press, 1992
   (chapter 7 "Random numbers", ran2 random number generator) */
{
	int j;
	long k;
	static long idum2=123456789;
	static long iy=0;
	static long iv[NTAB];
	double temp;

	if (idum <= 0) {
		if (-(idum) < 1)
			idum=1;
		else
			idum=-(idum);
		idum2=(idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(idum)/IQ1;
			idum=IA1*(idum-k*IQ1)-k*IR1;
			if (idum < 0)
				idum += IM1;
			if (j < NTAB)
				iv[j] = idum;
		}
		iy=iv[0];
	}
	k=(idum)/IQ1;
	idum=IA1*(idum-k*IQ1)-k*IR1;
	if (idum < 0)
		idum += IM1;
	k=idum2/IQ2;
	idum2=IA2*(idum2-k*IQ2)-k*IR2;
	if (idum2 < 0)
		idum2 += IM2;
	j=iy/NDIV;
	iy=iv[j]-idum2;
	iv[j] = idum;
	if (iy < 1)
		iy += IMM1;
	if ((temp=AM*iy) > RNMX)
		return RNMX;
	else
		return temp;
}

#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

int initrandom(int seed)
{
   srand((unsigned) time(NULL));
   if (seed < 0) 
	seed = rand();
   idum=-(long) seed;
   return (seed);
}  /* initrandom */ 


/* returns a random integer in the range [0; n - 1] */
int randominteger(int n)
{
	return (int) floor(randomunitintervall()*n);
}



#ifndef WITHOUT
int ct_sortkeyaddr(int addr)
{
  int a, res;
  a = addr % 4;
  res = addr - a + 3;
  return res;
}


int ct_nextedgeaddr(int addr)
{
  int a, res = 0;
  a = addr % 4;
  if ( a == 2 ) { res = addr - 2; }
  else          { res = addr + 1; }
  return res;
}

int ct_1stedge(int node)
{
  int res;
  res = 4 * node;
  return res;
}

int ct_2ndedge(int node)
{
  int res;
  res = 4 * node +1;
  return res;
}

int ct_3rdedge(int node)
{
  int res;
  res = 4 * node +2;
  return res;
}

int ct_isleaf(int node, int *ctree)
{
  return (ctree[ct_3rdedge(node)] < 0);
}

int ct_addr2node(int addr)
{
  int a, res;
  a = addr % 4;
  res = (int) ((addr - a) / 4);
  return res;
}

void fprintftreegraph(FILE *fp, int *ctree)
{
	int n;
	for (n=0; n < 2*Maxspc; n++) {
		fprintf(fp, "n[%3d] = (%3d.%2d, %3d.%2d, %3d.%2d | %3d)\n", n,
		(int) ctree[ct_1stedge(n)]/4,
		(int) ctree[ct_1stedge(n)]%4,
		(int) ctree[ct_2ndedge(n)]/4,
		(int) ctree[ct_2ndedge(n)]%4,
		(int) ctree[ct_3rdedge(n)]/4,
		(int) ctree[ct_3rdedge(n)]%4,
		ctree[ct_3rdedge(n)+1]);
	}
        fprintf(fp, "\n");
} /* fprintftreegraph */



int *inittree()
{
  int *snodes;
  int n;

  snodes = (int *) malloc(4 * 2 * Maxspc * sizeof(int));
  if (snodes == NULL) fprintf(stderr, "ERROR allocating snodes in inittree\n");

  for (n=0; n<(4 * 2 * Maxspc); n++) {
      snodes[n]=-1;
  }
  return snodes;
} /* inittree */


int *copytree(int *intree)
{
  int *outtree;

  outtree = malloc(4 * 2 * Maxspc * sizeof(int));
  if (outtree == NULL) fprintf(stderr, "ERROR allocating snodes in copytree\n");
  memcpy(outtree, intree, 4 * 2 * Maxspc * sizeof(int));
  return (outtree);

} /* copytree */


void freetree(int **snodes)
{
	free(*snodes);
	*snodes = NULL;
} /* freectree */


void copyOTU(int *ctree, int *ct_nextnode, int ct_curredge, int *ct_nextleaf, int ed)
{
#if 0
        int i, nextcurredge;

        /* test whether we are on a leaf */
        if (edge[ed].downright == NULL && edge[ed].downleft == NULL) {
                for (i = 1; i < nextleaf; i++) {
                        if (edgeofleaf[i] == ed) { /* i is the leaf of ed */
				nextcurredge          = ct_1stedge(*ct_nextleaf);
				ctree[ct_curredge]    = nextcurredge;
				ctree[nextcurredge]   = ct_curredge;
                                ctree[ct_sortkeyaddr(nextcurredge)] = i;
				(*ct_nextleaf)++;
                                return;
                        }
                }
        }

        /* we are NOT on a leaf */
	nextcurredge        = ct_1stedge(*ct_nextnode);
        ctree[ct_curredge]     = nextcurredge;
	ctree[nextcurredge] = ct_curredge;
        (*ct_nextnode)++;
	nextcurredge = ct_nextedgeaddr(nextcurredge);
        copyOTU(ctree, ct_nextnode, nextcurredge, ct_nextleaf, edge[ed].downleft->numedge);

	nextcurredge = ct_nextedgeaddr(nextcurredge);
        copyOTU(ctree, ct_nextnode, nextcurredge, ct_nextleaf, edge[ed].downright->numedge);
#endif
} /* copyOTU */


int sortOTU(int edge, int *ctree)
{
	int key1, key2;
	int edge1, edge2;
	int tempedge;

	if (ctree[ct_2ndedge((int) (edge / 4))] < 0)
		return ctree[ct_sortkeyaddr(edge)];

	edge1 = ctree[ct_nextedgeaddr(edge)];
	edge2 = ctree[ct_nextedgeaddr(ct_nextedgeaddr(edge))];

	key1  = sortOTU(edge1, ctree); 
	key2  = sortOTU(edge2, ctree); 
	
	if (key2 < key1) {
		tempedge            = ctree[ctree[edge1]];
		ctree[ctree[edge1]] = ctree[ctree[edge2]];
		ctree[ctree[edge2]] = tempedge;
		tempedge            = ctree[edge1];
		ctree[edge1]        = ctree[edge2];
		ctree[edge2]        = tempedge;
	  	ctree[ct_sortkeyaddr(edge)] = key2;
		
	} else {
	  ctree[ct_sortkeyaddr(edge)] = key1;
	}
	return ctree[ct_sortkeyaddr(edge)];
} /* sortOTU */

int sorttree(int *ctree, int startnode)
{
	int n;
	if ((startnode<0) || (startnode >= Maxspc)) {
        	for(n=0; n<Maxspc; n++) {
        	        if (ctree[ct_sortkeyaddr(n*4)] == 0)
        	                startnode = n;
        	}
	}
	sortOTU(ctree[startnode * 4], ctree);
	return startnode;
} /* sorttree */

void fprintfsortOTU(FILE *fp, int edge, int *ctree)
{
        int edge1, edge2;

        if (ctree[ct_2ndedge((int) (edge / 4))] < 0) {
                fprintf(fp, "%d", ctree[ct_sortkeyaddr(edge)]);
                return;
        }

        edge1 = ctree[ct_nextedgeaddr(edge)];
        edge2 = ctree[ct_nextedgeaddr(ct_nextedgeaddr(edge))];

        fprintf(fp, "(");
        fprintfsortOTU(fp, edge1, ctree); 
        fprintf(fp, ",");
        fprintfsortOTU(fp, edge2, ctree); 
        fprintf(fp, ")");

}


int fprintftree(FILE *fp, int *ctree, int startnode)
{
        int n; 
	if ((startnode<0) || (startnode >= Maxspc)) {
        	for(n=0; n<Maxspc; n++) {
        	        if (ctree[ct_sortkeyaddr(n*4)] == 0)
        	                startnode = n;
        	}
	}
        fprintf (fp, "(%d,", ctree[ct_sortkeyaddr(startnode*4)]);
        fprintfsortOTU(fp, ctree[startnode * 4], ctree);
        fprintf (fp, ");\n");
        return startnode;
} /* fprintftree */





void sprintfOTU(char *str, int *len, int edge, int *ctree)
{
        int edge1, edge2;

        if (ctree[ct_2ndedge((int) (edge / 4))] < 0) {
                *len+=sprintf(&(str[*len]), "%d", ctree[ct_sortkeyaddr(edge)]);
		return;
	}

        edge1 = ctree[ct_nextedgeaddr(edge)];
        edge2 = ctree[ct_nextedgeaddr(ct_nextedgeaddr(edge))];

	sprintf(&(str[*len]), "(");
	(*len)++;
        sprintfOTU(str, len, edge1, ctree); 
	sprintf(&(str[*len]), ",");
	(*len)++;
        sprintfOTU(str, len, edge2, ctree); 
	sprintf(&(str[*len]), ")");
	(*len)++;
} /* sprintfOTU */

char *sprintftree(int *ctree, int startnode, int strglen)
{
	char *treestr,
	     *tmpptr;
        int n,
	    len=0;
	treestr = (char *) malloc(strglen * sizeof(char));
	tmpptr  = treestr;

	if ((startnode<0) || (startnode >= Maxspc)) {
        	for(n=0; n<Maxspc; n++) {
        	        if (ctree[ct_sortkeyaddr(n*4)] == 0)
        	                startnode = n;
        	}
	}

	len+=sprintf (&(tmpptr[len]), "(%d,", ctree[ct_sortkeyaddr(startnode*4)]);
        sprintfOTU(tmpptr, &len, ctree[startnode * 4], ctree);
	len+=sprintf (&(tmpptr[len]), ");");
        return treestr;
}  /* sprintftree */


void fprintffulltree(FILE *outf, char *treestr, char **names)
{
	int count = 0;
	int idnum = 0;
	int n;
	for(n=0; treestr[n] != '\0'; n++){
		while(isdigit((int)treestr[n])){
			idnum = (10 * idnum) + ((int)treestr[n]-48);
			n++;
			count++;
		}
		if (count > 0){
			fprintf(outf, "'%s'", names[idnum]);
			count = 0;
			idnum = 0;
		}
		fprintf(outf, "%c", treestr[n]);
	}
}



void fprintfpossibleaddedges(FILE *fp, int *ctree, int ct_nextleaf, int ct_nextnode, int Maxspc)
{
	int n;
	for (n=0; n<(2 * Maxspc); n++) {
		if ((n < ct_nextleaf) && (n == 0)) {
			fprintf(fp, "0.0 (%d, root edge) -> %d.%d\n", 
				ct_1stedge(n), (int)(ctree[ct_1stedge(n)] / 4), 
				(int)(ctree[ct_1stedge(n)] % 4));
		}
		if ((n >= Maxspc) && (n < ct_nextnode)) {
			fprintf(fp, "%d.%d (%d) -> %d.%d\n", n, 
				(int)(ct_2ndedge(n) % 4), ct_2ndedge(n), 
				(int)(ctree[ct_2ndedge(n)] / 4), 
				(int)(ctree[ct_2ndedge(n)] % 4));
			fprintf(fp, "%d.%d (%d) -> %d.%d\n", n, 
				(int)(ct_3rdedge(n) % 4), ct_3rdedge(n), 
				(int)(ctree[ct_3rdedge(n)] / 4), 
				(int)(ctree[ct_3rdedge(n)] % 4));
		}
	}
}


void randtree(int *ctree)
{
        int ct_curredge;
        int ct_nextleaf;
        int ct_nextnode;
	int addedge;
	int tmpleaf;
	int tmpnode;

        ct_nextnode = Maxspc;
        ct_curredge = ct_1stedge(ct_nextnode);
        ct_nextleaf = 0;

        ctree[ct_1stedge(0)] = ct_curredge;
        ctree[ct_curredge]   = ct_1stedge(0);
        ctree[ct_sortkeyaddr(0)] = 0;
	ct_curredge = ct_nextedgeaddr(ct_curredge);
	ct_nextleaf++;

        ctree[ct_1stedge(2)] = ct_curredge;
        ctree[ct_curredge]   = ct_1stedge(2);
        ctree[ct_sortkeyaddr(ct_1stedge(2))] = 2;
	ct_curredge = ct_nextedgeaddr(ct_curredge);
	ct_nextleaf++;

        ctree[ct_1stedge(1)] = ct_curredge;
        ctree[ct_curredge]   = ct_1stedge(1);
        ctree[ct_sortkeyaddr(ct_1stedge(1))] = 1;
	ct_curredge = ct_nextedgeaddr(ct_curredge);
	ct_nextleaf++;

        ct_nextnode++;
        
	for (tmpleaf=ct_nextleaf; tmpleaf< Maxspc; tmpleaf++){
		tmpnode = ct_nextnode++;
		ct_nextleaf++;
		ctree[ct_2ndedge(tmpnode)] = ct_1stedge(tmpleaf);
		ctree[ct_1stedge(tmpleaf)] = ct_2ndedge(tmpnode);
		ctree[ct_sortkeyaddr(ct_1stedge(tmpleaf))]=tmpleaf;

		/* add to edge */
		{int num;
		 num = randominteger((ct_nextnode - Maxspc - 1) * 2 + 1);
		 if (num == 0) {
			addedge = ct_1stedge(0);
		 } else {
			if (((num-1) % 2) == 0) {
				addedge = ct_2ndedge((int)((num-1) / 2) + Maxspc);
			} else {
				addedge = ct_3rdedge((int)((num-1) / 2) + Maxspc);
			}
		 }
		}

		ctree[ct_1stedge(tmpnode)] = addedge;
		ctree[ct_3rdedge(tmpnode)] = ctree[addedge];
		ctree[addedge] = ct_1stedge(tmpnode);
		ctree[ctree[ct_3rdedge(tmpnode)]] = ct_3rdedge(tmpnode);
		
	}
        ct_curredge = ct_nextedgeaddr(ct_curredge);

        ct_curredge = ct_nextedgeaddr(ct_curredge);
} /* randtrees */


void addrekursive(int *ctree, 
                  int  ct_nextleaf, 
                  int  ct_nextnode)
{
	int addedge;
	int tmpleaf;
	int tmpnode;
	int num;
	int *cptr;

        
	if (ct_nextleaf < Maxspc) {
		tmpleaf = ct_nextleaf++; 
		tmpnode = ct_nextnode++;

		ctree[ct_2ndedge(tmpnode)] = ct_1stedge(tmpleaf);
		ctree[ct_1stedge(tmpleaf)] = ct_2ndedge(tmpnode);
		ctree[ct_sortkeyaddr(ct_1stedge(tmpleaf))]=tmpleaf;

		/* add to edge */
		for (num=0; num <((ct_nextnode - Maxspc - 1) * 2 + 1); num++) {
			cptr = copytree(ctree);
			if (num == 0) 
			   addedge = ct_1stedge(0);
			else {
			   if (((num-1) % 2) == 0)
			      addedge = ct_2ndedge((int)((num-1) / 2) + Maxspc);
			   else
			      addedge = ct_3rdedge((int)((num-1) / 2) + Maxspc);
			}
			cptr[ct_1stedge(tmpnode)] = addedge;
			cptr[ct_3rdedge(tmpnode)] = cptr[addedge];
			cptr[addedge] = ct_1stedge(tmpnode);
			cptr[cptr[ct_3rdedge(tmpnode)]] = ct_3rdedge(tmpnode);
			addrekursive(cptr, ct_nextleaf, ct_nextnode);
			freetree(&cptr);
		}
	} else {
		sorttree(ctree,0);
		treestr = sprintftree(ctree, 0, treestrlen);
		fprintffulltree(outfile, treestr, spcnames);
		free(treestr);
		fprintf(outfile, "\n");
		printdots(stderr, &dotnum, 10, '.', 100, ' ', 500, '\n');
	}
} /* addrekursive */


void gentrees()
{
        int ct_curredge;
        int ct_nextleaf;
        int ct_nextnode;
	int *ctree;

	ctree = inittree();

        ct_nextnode = Maxspc;
        ct_curredge = ct_1stedge(ct_nextnode);
        ct_nextleaf = 0;

        ctree[ct_1stedge(0)] = ct_curredge;
        ctree[ct_curredge]   = ct_1stedge(0);
        ctree[ct_sortkeyaddr(0)] = 0;
	ct_curredge = ct_nextedgeaddr(ct_curredge);
	ct_nextleaf++;

        ctree[ct_1stedge(2)] = ct_curredge;
        ctree[ct_curredge]   = ct_1stedge(2);
        ctree[ct_sortkeyaddr(ct_1stedge(2))] = 2;
	ct_curredge = ct_nextedgeaddr(ct_curredge);
	ct_nextleaf++;

        ctree[ct_1stedge(1)] = ct_curredge;
        ctree[ct_curredge]   = ct_1stedge(1);
        ctree[ct_sortkeyaddr(ct_1stedge(1))] = 1;
	ct_curredge = ct_nextedgeaddr(ct_curredge);
	ct_nextleaf++;

        ct_nextnode++;
        
	addrekursive(ctree, ct_nextleaf, ct_nextnode);
	freetree(&ctree);

} /* gentrees */

#endif /* WITHOUT */

char **getnamefromfile(char *filename, int *maxspc)
{
#	define MAXLINELEN 10000
	FILE *fp;
	int numspc = 0;
	char **nameptr;
	char dummy[MAXLINELEN];
	int slen;
	fp = fopen(filename, "r");
	if (fp == NULL) {
		fprintf(stderr, "Can't open file: %s\n", filename);
		exit (2);
	}
	fgets(dummy, MAXLINELEN, fp);
	sscanf(dummy, "%d", &numspc);
	if (numspc < 3) {
		fprintf(stderr, "Number of species names must be at least 3\n");
		exit (3);
	}
	*maxspc = numspc;

	fprintf(stderr, "Reading %d species names from file '%s':\n\n", numspc, filename);
	nameptr = (char **)malloc(*maxspc * sizeof(char *));
	for (numspc = 0; numspc < *maxspc; numspc++) {
		nameptr[numspc] = (char *)malloc(11 * sizeof(char));
		fgets(dummy, MAXLINELEN, fp);
		sscanf(dummy, "%10c", nameptr[numspc]);
		slen = strlen(nameptr[numspc]) - 1; 
		while ((slen >= 0) 
			&& ((nameptr[numspc][slen] == ' ') 
			   || (nameptr[numspc][slen] == '\t')
			   || (nameptr[numspc][slen] == '\n'))) {
			nameptr[numspc][slen] = '\0';
			slen--;
		}
		fprintf(stderr, "%3d. \'%s\'\n", numspc+1, nameptr[numspc]);
	}
	fclose(fp);
	return (nameptr);
} /* getnamefromfile */


void printusage(int argc, char *argv[])
{
	fprintf(stderr, "\nUsage: %s <#samples> <specfile> [<outfile>]\n", argv[0]);
	fprintf(stderr, "\n   #samples - Number of random tree samples");
	fprintf(stderr, "\n   specfile - File with species name (Phylip interleaved alignment format)");
	fprintf(stderr, "\n   outfile  - Output file, if not given trees are printed to stdout\n\n");
}




int main(int argc, char *argv[])
{
	int *tree;
	int  n;
	int  numtrees;
	int  numedges;
	int  maxsamp;

	if ((argc <= 2) || (argc > 4)) {
		printusage(argc, argv);
		exit (1);
	}
	fprintf(stderr, "\n");
	maxsamp = atoi(argv[1]);
	spcnames = getnamefromfile(argv[2], &Maxspc);
	if (argc == 4) {
		outfile = fopen(argv[3], "w");
		if (outfile == NULL) {
			fprintf(stderr, "Can't open file to write: %s\n", argv[3]);
			exit(4);
		}
	} else
		outfile = stdout;
	treestrlen = (Maxspc * (int)(1 + log10(Maxspc))) + (Maxspc * 3);
	
	if (maxsamp == 0) {
		numtrees = 1;
		numedges = 3;
		for (n=3; n<Maxspc; n++) {
			numtrees *= numedges; 
			numedges += 2;
		}
		fprintf(stderr, "\nGenerating all %d possible binary trees (1 dot per 10 trees):\n\n", numtrees);
		gentrees();
		fprintf(stderr, "\n\n%d trees generated\n\n", dotnum);
	} else {
        	(void) initrandom(-1);
		fprintf(stderr, "\nGenerating %d random sample trees (1 dot per 10 trees):\n\n", maxsamp);
		for (n=0; n<maxsamp; n++) {
			tree = inittree();
			randtree(tree);
			sorttree(tree,0);
	
			treestr = sprintftree(tree, 0, treestrlen);
			fprintffulltree(outfile, treestr, spcnames);
			free(treestr);
			fprintf(outfile, "\n");
			freetree(&tree);
			printdots(stderr, &dotnum, 10, '.', 100, ' ', 500, '\n');
		} /* for n */
		fprintf(stderr, "\n\n%d trees generated\n\n", dotnum);
	} /* if maxsamp == 0 else */
	return (0);
} /* main */
